#ifndef DYN_2D_HPP
#define DYN_2D_HPP

#include <iostream>
#include <initializer_list>
#include "empty_creature.hpp"
#include "creature.hpp"
#include "random_gen.hpp"
#include "define_config.hpp"

namespace ecosystem{
	
	
	template <typename T>
	class Dynamic2D{
		private:
			T** in_array=nullptr;
			std::size_t width=0;
			std::size_t height=0;
		public:
			Dynamic2D()=default;
			
			Dynamic2D(std::size_t awidth,std::size_t aheight, T def_value){
				resize(awidth,aheight,def_value);
			}
			Dynamic2D(std::initializer_list<std::initializer_list<T>> values) {
				height = values.size();
				width = values.begin()->size();
				in_array = new T*[height];
		
				std::size_t i = 0;
				for (const auto& row : values) {
					in_array[i] = new T[width];
					std::copy(row.begin(), row.end(), in_array[i]); //copy row values
					i++;
				}
			}
			
			void swap_at(int x1,int y1, int x2, int y2){
				
				std::swap(in_array[y1][x1], in_array[y2][x2]);

				if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=4){
					std::cout << "Swapped X1:" <<x1 << " Y1:" << y1 << " with X2:" << x2 << " Y2:" << y2 <<".\n";
				}
			}
			
			bool is_valid_coordinate(int x,int y){
				if(x<0||y<0||x>=width||y>=height){
					return false;
				}
				return true;
			}
			
			T*& operator[](std::size_t row) {
				return in_array[row];
			}
			T* const& operator[](std::size_t row) const {
				return in_array[row];
			}
			
			void print(){
				for(std::size_t y=0;y<height;y++){
					for(std::size_t x=0;x<width;x++){
						#ifdef UNICODE_PRINT_SUPPORT
							std::wcout << "|" << in_array[y][x]->get_unisymbol();
						#else
							std::cout << "|" << *in_array[y][x];
						#endif
					}
					std::cout << "|\n";
				}
			}
			void reset_states(){
				for(std::size_t y=0;y<height;y++){
					for(std::size_t x=0;x<width;x++){
						in_array[y][x]->updated=false;
					}
				}
			}
			void update_loop(){
				for(std::size_t y=0;y<height;y++){
					for(std::size_t x=0;x<width;x++){
						if(in_array[y][x]->updated==false){
							in_array[y][x]->update(*this,x,y);
						}
					}
				}
			}
			void cleanup_loop(){
				for(std::size_t y=0;y<height;y++){
					for(std::size_t x=0;x<width;x++){
						if(in_array[y][x]->get_living_state()==Livingstate::DEAD){
							if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=3){
								std::cout << *in_array[y][x] << " died at X:" << x << " Y:" << y << ". Cleaning up...\n";
							}
							delete in_array[y][x];
							in_array[y][x]=new Empty();
						}
					}
				}
			}
			void replace_at(int x, int y, T new_data){
				if(in_array[y][x]){
					delete in_array[y][x];
				}
				in_array[y][x]=new_data;
			}
			
			void clear(){
				for(std::size_t i=0;i<height;i++){
					delete[] in_array[i];
				}
				delete[] in_array;
			}
			void resize(std::size_t awidth,std::size_t aheight, T def_value){
				if(in_array!=nullptr){
					clear();
				}
				
				width=awidth;
				height=aheight;
				
				in_array = new T*[height];
				for(std::size_t i=0;i<height;i++){
					in_array[i]=new T[width];
					
					std::fill(in_array[i], in_array[i] + width, def_value);
				}
			}
			
			~Dynamic2D(){
				clear();
			}
	};
}
	
#endif